#ifndef ERL_STRING
#define ERL_STRING 1
#define STRING_QUANT 32
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
extern "C" {
    char    *MallocString (int size) ;
    void    FreeString (char *string, int length) ;
}

class EString {
    
    public :
    
        EString ()
            : pvLength(0),  pvStart(0)
        {
            Malloc(0);
            *pvString = '\0';
        }
        
        EString ( const char *val, int siz )
            : pvStart(0)
        {
            if ( !val ) 
                siz = 0 ;
            Malloc(siz + 1);
            pvLength = siz ;
            if ( siz ) 
                memcpy(pvString, val, pvLength + 1);
            *(pvString + siz) = '\0';
        }
        
        EString ( const char *val )
            : pvStart(0)
        {
            if ( !val ) 
                pvLength = 0 ;
            else 
                pvLength = strlen(val);
            Malloc(pvLength + 1);
            if ( pvLength ) 
                memcpy(pvString, val, pvLength + 1);
            else 
                *pvString = '\0';
        }
        
        EString ( const EString &str )
            : pvStart(0)
        {
            Malloc((pvLength = str.length()) + 1);
            memcpy(pvString, str.c_str(), pvLength + 1);
        }
        
        EString ( char c )
            : pvStart(0)
        {
            Malloc((pvLength = 1) + 1);
            *pvString = c ;
            *(pvString+1) = '\0';
        }
        
        ~EString ()
        {
            Free();
        }
        
        EString &operator= ( const EString &stlString )
        {
            if ( this == &stlString ) 
                return *this ;
            if ( pvSize < stlString.length() + 1 ) {
                Realloc((pvLength = stlString.pvLength) + 1
                    , stlString.c_str(), stlString.pvLength + 1);
                pvStart = 0 ;
                return *this ;
            } else if ( stlString.pvLength < STRING_QUANT
                            && pvSize > 6 * STRING_QUANT ) {
                
                // limit growth of string
                Realloc(6 * STRING_QUANT, 0, 0);
            }
            pvStart = 0 ;
            memcpy((char *)c_str(), stlString.c_str()
                , (pvLength = stlString.pvLength) + 1);
            return *this ;
        }
        
        EString &operator= ( const char *str )
        {
            if ( !str ) 
                pvLength = 0 ;
            else 
                pvLength = strlen(str);
            if ( pvSize < pvLength + 1 ) {
                Realloc(pvLength + 1, str, pvLength);
                pvStart = 0 ;
            } else if ( pvLength < STRING_QUANT
                            && pvSize > 6 * STRING_QUANT ) {
                
                // limit growth of string
                // limit growth of string
                Realloc(6 * STRING_QUANT, 0, 0);
            }
            pvStart = 0 ;
            if ( pvLength ) 
                memcpy((char *)c_str(), str, pvLength + 1);
            else 
                *((char *)(c_str())) = '\0';
            return *this ;
        }
        
        EString &operator= ( char c )
        {
            pvLength = 1 ;
            if ( pvSize < 2 ) {
                Realloc(2, 0, 0);
            } else if ( pvSize > 6 * STRING_QUANT ) {
                
                // limit growth of string
                Realloc(6 * STRING_QUANT, 0, 0);
            }
            pvStart = 0 ;
            *((char *)c_str()) = c ;
            *((char *)c_str() + 1) = '\0';
            return *this ;
        }
        
        EString &prepend ( const char *val )
        {
            unsigned int    length ;
            
            if ( !val ) 
                return *this ;
            length = strlen(val);
            if ( pvStart >= length ) {
                memcpy((char *)c_str() - length, val, length);
                pvStart -= length ;
                pvLength += length ;
            } else {
                Realloc(length + pvLength + 1, pvString, - (pvLength + 1));
                pvStart = pvSize - pvLength - 1 - length ;
                memcpy((char *)c_str(), val, length);
                pvLength += length ;
            }
            *((char *) c_str() +pvLength) = '\0';
            return *this ;
        }
        
        EString at ( unsigned int deb, unsigned int len ) const
        {
            if ( deb >= pvLength ) 
                deb = pvLength ;
            if ( deb + len > pvLength ) 
                len = pvLength - deb ;
            return EString(c_str() + deb, len);
        }
        
        EString substring ( int deb, int len ) const
        {
            return at(deb, len);
        }
        
        EString &operator+= ( const char *str )
        {
            int length ;
            
            if ( !str ) 
                return *this ;
            if ( pvLength + (length = strlen(str)) + 1 > pvSize - pvStart ) {
                Realloc(pvLength + length + 1, pvString, pvLength);
                pvStart = 0 ;
            }
            if ( length == 1 ) {
                char    *ptStr = (char *)c_str() + pvLength ;
                *ptStr++ = *str ;
                *ptStr = '\0';
                pvLength++ ;
            } else {
                memcpy((char *)c_str() + pvLength, str, length + 1);
                pvLength += length ;
            }
            return *this ;
        }
        
        EString &operator+= ( const EString &str )
        {
            int length = str.length();
            
            if ( pvLength + length + 1 > pvSize - pvStart ) {
                Realloc(pvLength + length + 1, pvString, pvLength);
                pvStart = 0 ;
            }
            memcpy((char *)c_str() + pvLength, str.c_str(), length + 1);
            pvLength += length ;
            return *this ;
        }
        
        EString &operator+= ( const char c )
        {
            int length ;
            
            if ( pvLength + 2 >= pvSize - pvStart ) {
                Realloc(pvLength + 2, pvString, pvLength);
                pvStart = 0 ;
            }
            *((char *)c_str() + pvLength) = c ;
            ++pvLength ;
            *((char *)c_str() + pvLength) = '\0';
            return *this ;
        }
        
        bool operator== ( const char *str ) const
        {
            unsigned int    length ;
            
            if ( !str ) 
                return pvLength == 0 ;
            length = strlen(str);
            if ( pvLength == length ) {
                return !strncmp(c_str(), str, length);
            }
            return false ;
        }
        
        EString operator+ ( const char *str ) const
        {
            EString result(*this);
            
            result += str ;
            return EString(result);
        }
        
        EString operator+ ( const char c ) const
        {
            EString result(*this);
            
            result += c ;
            return EString(result);
        }
        
        EString operator+ ( const EString &str ) const
        {
            EString result(*this);
            
            result += str.c_str();
            return EString(result);
        }
        
        bool operator!= ( const char *str ) const
        {
            return !operator== (str);
        }
        
        bool operator== ( const EString &str ) const
        {
            if ( pvLength == str.length() ) {
                return !strncmp(c_str(), str.c_str(), pvLength);
            }
            return false ;
        }
        
        bool operator!= ( const EString &str ) const
        {
            return !operator== (str);
        }
        
        const char *c_str () const
        {
            return pvString + pvStart ;
        }
        
        operator const char *() const
        {
            return c_str();
        }
        
        char operator[] ( unsigned int pos ) const
        {
            if ( pos < pvLength ) 
                return *(c_str() + pos);
            return '\0';
        }
        
        const unsigned int length () const
        {
            return pvLength ;
        }
        
        bool operator< ( const EString &str ) const
        {
            unsigned int    shtLength = length(); // length of shorter string
            
            if ( str.length() < shtLength ) {
                shtLength = str.length();
            }
            
            int comp ; // result of comparison
            
            // if different return result
            if ( comp = strncmp(c_str(), str.c_str(), shtLength) ) {
                return comp < 0 ;
            } else {
                
                // true if only the first is shorter
                return length() == shtLength && shtLength != str.length();
            }
        }
        
        bool operator> ( const EString &str ) const 
        {
            unsigned int    shtLength = length(); // length of shorter string
            
            if ( str.length() < shtLength ) {
                shtLength = str.length();
            }
            
            int comp ; // result of comparison
            
            // if different return result
            if ( comp = strncmp(c_str(), str.c_str(), shtLength) ) {
                return comp > 0 ;
            } else {
                
                // true if only the first is longer
                return length() != shtLength ;
            }
        }
        
        virtual void Free ()
        {
            FreeString(pvString, pvSize - 1);
        }
        
        // Malloc : malloc new space for string
        // parameters : 
        //              size : new size
        virtual void Malloc ( int size )
        {
            pvString = MallocString(pvSize = size + STRING_QUANT);
        }
        
        // Realloc : realloc the string
        //           size : new size
        //           paramNewString : string to be copied inside 
        //           lght : store it starting from end if not 0
        virtual void Realloc ( int size, const char *paramNewString
            , int lght )
        {
            
            char    *newString ;   // new string
            int     bigOffset(0);  // offset for big strings
            
            if ( size > 300 * STRING_QUANT ) {
                bigOffset = 300 * STRING_QUANT ;
            }
            
            // create a copy of stlString
            newString = MallocString(size + STRING_QUANT + bigOffset);
            if ( paramNewString ) 
                if ( lght > 0 ) 
                    memmove(newString, paramNewString, lght);
                else if (lght <0)
                    memmove(
                        newString + STRING_QUANT + size + bigOffset + lght
                        , paramNewString, - lght);
            FreeString((char *)c_str(), pvSize - 1);
            
            // insert copy here
            pvString = newString ;
            pvSize = size + STRING_QUANT + bigOffset ;
        }
        
        // Resize : resize string 
        //          size : new size
        void Resize ( unsigned int size )
        {
            if ( size + STRING_QUANT < pvLength + 1 ) {
                if ( pvLength + 1 > STRING_QUANT ) {
                    size = pvLength + 1 - STRING_QUANT ;
                } else 
                    size = 1 ;
            }
            Realloc(size, pvString, pvLength + 1);
        }
        
        // Size : gives size
        // return : size of string
        unsigned int Size ()
        {
            return pvSize ;
        }
    
    private :
    
        char            *pvString ;
        unsigned int    pvSize ;
        unsigned int    pvLength ;
        unsigned int    pvStart ;
};

class LessString {
    
    public :
    
        LessString () {}
        
        ~LessString () {}
        
        bool operator() ( const EString &x, const EString &y ) const
        {
            return ((EString)x).EString::operator< (y);
        }
};
#endif
