 language metachop;

static  int GetNumber ( const PTREE &tree )
{
    return APPLY_CLASS(tree, TreeClass, Type());
}

#include "token.h"

// EncodeClass : encoding of a class
// parameters :
//              tree : the tree to be encoded
//              string : recipient string
void PTREE::EncodeClass ( const PTREE &tree, EString &string )
{
    char    *pt ; // pointer on string
    
    CoarseEncodeInt(GetNumber(tree), string);
    pt = (char *)(tree.Value());
    CoarseEncodeInt(strlen(pt), string);
    string += pt ;
    return ;
}

// CoarseEncode : CoarseEncode a tree
// parameters :
//              tree : the tree to be encoded
//              string : recipient string
void PTREE::CoarseEncode ( const PTREE &tree, EString &string )
{
    
    int     arity ;  // arity of node
    int     nbNode ; // number of node
    char    *pt ;    // a pointer
    
start : 
    if ( !tree ) {
        string += (char) - 1 ;
        return ;
    }
    switch ( tree.NumberTree() ) {
        case TERM_TREE : 
            string += '\1';
            CoarseEncodeInt(TERM_TREE, string);
            pt = (char *)(tree.Value());
            CoarseEncodeInt(strlen(pt), string);
            string += pt ;
			CoarseEncode(tree[0],string);
            return ;
        case CLASS_TREE : 
            string += '\1';
            CoarseEncodeInt(CLASS_TREE, string);
            EncodeClass(tree, string);
			CoarseEncode(tree[0],string);
            return ;
        case REF_TREE : 
            string += (char) - 1 ;
            return ;
        default : 
            {
                
                /* a tree node */
                arity = treearity(tree);
                string += (char)arity ;
                CoarseEncodeInt(tree.NumberTree(), string);
                
                /* put good language mask */
                int indexSon ; // index of son
                for ( indexSon = 1 ; indexSon <= arity ; indexSon++ ) {
                    CoarseEncode(tree [indexSon], string);
                }
				CoarseEncode(tree[0],string);
            }
    }
}

// DecodeClass : Decode a class 
// parameters :
//              string : string to be decoded
// return : the tree
PTREE PTREE::DecodeClass ( char *&string )
{
    
    PTREE   myTree ;  // return tree
    PTREE   topTree ; // top  tree
    int     nbNode ;  // number of node
    int     length ;  // length of string
    
    nbNode = CoarseDecodeInt(string);
    length = CoarseDecodeInt(string);
    myTree = LMakeString(string, length);
    topTree = PTREE(nbNode, 1);
    topTree.ReplaceTree(1, myTree);
    string += length ;
    return topTree ;
}

// CoarseDecode : CoarseDecode a tree
// parameters :
//              string : input string
// return : the tree
PTREE PTREE::CoarseDecode ( char *&string )
{
    char    arity ;
    PTREE   myTree ;
    int     nbNode ;
    int     length ;
    
    arity = *string++ ;
    
    // get arity
    if ( arity == (char) - 1 ) 
        return (PTREE)0 ;
    
    // get node
    nbNode = CoarseDecodeInt(string);
    
    // decode value
    switch ( arity ) {
        case 1 : 
            switch ( nbNode ) {
                case TERM_TREE : 
                    length = CoarseDecodeInt(string);
                    myTree = LMakeString(string, length);
                    string += length ;
                    myTree.ReplaceTree(0,CoarseDecode(string));
                    return myTree ;
                    break ;
                case CLASS_TREE : 
                    {
                        myTree = DecodeClass(string);
						myTree.ReplaceTree(0,CoarseDecode(string));
                    }
                    break ;
                default : 
                    myTree = PTREE(nbNode, (int)arity);
                    myTree.ReplaceTree(1, CoarseDecode(string));
                    myTree.ReplaceTree(0, CoarseDecode(string));
                    break ;
            }
            break ;
        default : 
            {
                
                // construct node
                myTree = PTREE(nbNode, (int)arity);
                
                // put sons
                int indexSon ; // index of son
                for ( indexSon = 1 ; indexSon <= arity ; indexSon++ ) {
                    myTree.ReplaceTree(indexSon, CoarseDecode(string));
                }
				myTree.ReplaceTree(0,CoarseDecode(string));
            }
    }
    return myTree ;
}
